/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Print information about other known peers.
 *
 * @author Christian Grothoff
 * @file server/gnunet-peer-info.c
 **/

#include <stdio.h>
#include <stdlib.h>
#include "server/transport.h"
#include "server/keyservice.h"
#include "server/knownhosts.h"
#include "util/logging.h"
#include "util/initialize.h"
#include "util/getopt.h"
#include "util/xmalloc.h"
#include "util/configuration.h"

/**
 * Perform option parsing from the command line. 
 **/
static int parser(int argc, 
	   char * argv[]) {
  int cont = OK;
  int c;


  freeNonNull(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL),
	      "gnunet-peer-check always use option d");
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      { "loglevel",1, 0, 'L' },
      { "config",  1, 0, 'c' },
      { "version", 0, 0, 'v' },
      { "help",    0, 0, 'h' },
      { 0,0,0,0 }
    };
    
    c = getopt_long(argc,
		    argv, 
		    "vhc:L:", 
		    long_options, 
		    &option_index);
    
    if (c == -1) 
      break;  // No more flags to process
    
    switch(c) {
    case 'c':
      freeNonNull(setConfigurationString("FILES",
					 "gnunet.conf",
					 optarg),
		  "startup: option parsing c"); 
      break;
    case 'v': 
      printf("gnunet-peer-info v%s\n",
	     VERSION);
      cont = SYSERR;
      break;
    case 'h': 
      printf("USAGE: gnunet-peer-info [OPTIONS]\n"\
	     "-h, --help                    : "
	     "print this page\n"
	     "-c FILENAME, --config=FILENAME: "
	     "load config file (defaults: %s)\n"
	     "-v, --version                 : "
	     "print the version number\n"\
	     "-L LOGLEVEL, --loglevel=LOGL  : "
	     "set the loglevel\n",
	     DEFAULT_CONFIG_FILE);
      cont = SYSERR;
      break;
    case 'L':
      freeNonNull(setConfigurationString("GNUNETD",
					 "LOGLEVEL",
					 optarg),
		  "startup option parsing L");
      break;
    default:
      LOG(LOG_FAILURE, 
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      cont = SYSERR;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    LOG(LOG_WARNING, 
	"WARNING: Invalid arguments: ");
    while (optind < argc)
      LOG(LOG_WARNING, 
	  "%s ", argv[optind++]);
    LOG(LOG_FATAL,
	"FATAL: Invalid arguments. Exiting.\n");
    return SYSERR;
  }
  return cont;
}

/* make linker happy... */
static CoreAPIForTransport capi;
CoreAPIForTransport * getCoreAPIForTransport() {
  return &capi;
}

/**
 * Print information about the peer.
 * Currently prints only the HostIdentity
 * and the IP. Could of course do more
 * (e.g. resolve via DNS or print trust).
 **/ 
static void printHostInfo(const HostIdentity * id,
			  const unsigned short proto,
			  void * data) {
  HELO_Message * helo;
  char * info;
  HexName hex;

  hash2hex(&id->hashPubKey,
	   &hex);
  if (SYSERR == identity2Helo(id,
			      proto,
			      NO,
			      &helo)) {
    LOG(LOG_WARNING,
	"WARNING: could not get address of %s\n",
	&hex);
    return;
  }  
  if (SYSERR == verifySig(&helo->senderIdentity,
			  HELO_Message_size(helo) - sizeof(Signature) - sizeof(PublicKey) - sizeof(p2p_HEADER),
			  &helo->signature,
			  &helo->publicKey)) {
    LOG(LOG_WARNING, 
	"WARNING: HELO message invalid (signature bogous).\n");
  }
  info = heloToString(helo);
  xfree(helo,
	"freeing helo");
  if (info == NULL) {
    LOG(LOG_WARNING,
	"WARNING: could not get address of %s\n",
	&hex);
    return;
  }
  printf("%s at address %s\n",
	 (char*)&hex,
	 info);
  xfree(info,
	"heloToString characters");  
}

int main(int argc, char *argv[]) {
  
  if (OK != initUtil(argc, argv, &parser))
    return SYSERR;

  initKeyService();
  initTransports();
  initKnownhosts();
  forEachHost(&printHostInfo,
	      0,
	      NULL);
  doneKnownhosts();
  doneTransports();
  doneKeyService();
  doneUtil();
  return 0;
}


/* end of gnunet-peer-info.c */
