/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Helper methods for the startup of gnunetd:
 * - install signal handling
 * - system checks on startup
 * - PID file handling
 * - detaching from terminal
 * - command line parsing
 *
 * @file server/startup.c
 * @author Christian Grothoff
 **/

#include "config.h"
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <signal.h>
#include "util/getopt.h"
#include "util/initialize.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "server/startup.h"
#include "server/transport.h"

extern int debug_flag;
/**
 * This flag is set if gnunetd is shutting down.
 **/
static Semaphore * doShutdown;

/* ************* SIGNAL HANDLING *********** */

/**
 * Cron job that triggers re-reading of the configuration.
 **/
static void reread_config_helper(void * unused) {
  LOG(LOG_MESSAGE,
      "Re-reading configuration\n");
  readConfiguration();
  resetUtil();
  LOG(LOG_MESSAGE,
      "Re-reading configuration done.\n");
}

/**
 * Signal handler for SIGHUP.
 * Re-reads the configuration file. 
 **/
static void reread_config(int signum) {
  LOG(LOG_DEBUG, 
      "DEBUG: Received SIGHUP %d\n",
      signum);
  addCronJob(&reread_config_helper,
	     1 * cronSECONDS, 
	     0,
	     NULL);
}

/**
 * Try a propper shutdown of gnunetd.
 **/
static void shutdown_gnunetd(int signum) {
  LOG(LOG_DEBUG, 
      "DEBUG: Received SIGTERM %d\n",
      signum);
  semaphore_up(doShutdown);
}

/**
 * Initialize signal handlers
 **/
void initSignalHandlers() {
  struct sigaction sig;
  struct sigaction oldsig;

  doShutdown = new_semaphore(0);
  sig.sa_handler = &shutdown_gnunetd;
  sigemptyset(&sig.sa_mask);
#ifdef SA_INTERRUPT
  sig.sa_flags = SA_INTERRUPT; /* SunOS */
#else
  sig.sa_flags = SA_RESTART;
#endif    
  sigaction(SIGINT,  &sig, &oldsig);
  sigaction(SIGTERM, &sig, &oldsig);
  sigaction(SIGQUIT, &sig, &oldsig);
  
  sig.sa_handler = &reread_config;
  sigaction(SIGHUP, &sig, &oldsig);
}

void doneSignalHandlers() {
  struct sigaction sig;
  struct sigaction oldsig;

  sig.sa_handler = SIG_DFL;
  sigemptyset(&sig.sa_mask);
#ifdef SA_INTERRUPT
  sig.sa_flags = SA_INTERRUPT; /* SunOS */
#else
  sig.sa_flags = SA_RESTART;
#endif
  sigaction(SIGINT,  &sig, &oldsig);
  sigaction(SIGTERM, &sig, &oldsig);
  sigaction(SIGQUIT, &sig, &oldsig);
  free_semaphore(doShutdown);  
}

void waitForSignalHandler() {
  semaphore_down(doShutdown);
}

/* *********** SYSTEM CHECKS ON STARTUP ************ */

/**
 * Check if the compiler did a decent job aligning the structs...
 **/
void checkCompiler() {
  if (sizeof(HELO_Message) != 556)
    errexit("sizeof HELO Message wrong! (%d != 556)",
	    sizeof(HELO_Message));
  if (sizeof(p2p_HEADER) != 4)
    errexit("sizeof p2p_HEADER wrong! (%d != 4)",
	    sizeof(p2p_HEADER));
}

/* *********** PID file handling *************** */

static char * getPIDFile() {
  return getFileName("GNUNETD",
		     "PIDFILE",
		     "You must specify a name for the PID file in section"\
		     " %s under %s\n");
}

/**
 * Write our process ID to the pid file.
 **/
void writePIDFile() {
  FILE * pidfd;
  char * pif;
  
  pif = getPIDFile();
  pidfd = fopen(pif, "w");
  if (pidfd == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: Could not write PID to file %s\n", 
	pif);
  } else {
    fprintf(pidfd, "%u", getpid());
    fclose(pidfd);
  }
  xfree(pif,
	"pid filename"); 
}

void deletePIDFile() {
  char * pif = getPIDFile();
  unlink(pif);    
  xfree(pif, "PID file name");
}

/* ************** DETACHING FROM TERMAINAL ************** */

/**
 * Fork and start a new session to go into the background
 * in the way a good deamon should.
 *
 * @param filedes pointer to an array of 2 file descriptors
 *        to complete the detachment protocol (handshake)
 **/
void detachFromTerminal(int * filedes) {
  pid_t pid;
  int nullfd;
  
   /* Don't hold the wrong FS mounted */
  if (chdir("/") < 0) {
    perror("chdir");
    exit(1);
  }

  pipe(filedes);
  pid = fork();
  if (pid < 0) {
    perror("fork");
    exit(1);
  }
  if (pid) {  /* Parent */
    int ok;
    char c;
    
    close(filedes[1]); /* we only read */
    ok = SYSERR;
    while (0 < read(filedes[0], &c, sizeof(char))) {
      if (c == '.') 
	ok = OK;
    }      
    fflush(stdout);
    if (ok == OK)
      exit(0);   
    else
      exit(1); /* child reported error */
  }
  close(filedes[0]); /* we only write */
  nullfd = open("/dev/null", O_CREAT | O_RDWR | O_APPEND);
  if (nullfd < 0) {
    perror("/dev/null");
    exit(1);
  }   
  /* child - close fds linking to invoking terminal, but
   * close usual incoming fds, but redirect them somewhere
   * useful so the fds don't get reallocated elsewhere.
   */
  if (dup2(nullfd,0) < 0 || 
      dup2(nullfd,1) < 0 || 
      dup2(nullfd,2) < 0) {
    perror("dup2"); /* Should never happen */
    exit(1);
  }
  pid = setsid(); /* Detach from controlling terminal */
}

void detachFromTerminalComplete(int * filedes) {
  char c = '.';
  write(filedes[1], &c, sizeof(char)); /* signal success */
  close(filedes[1]);
}

/* ****************** COMMAND LINE PARSING ********************* */

static void printDot(void * unused) {
  LOG(LOG_DEBUG, ".");
}

/**
 * Print a list of the options we offer.
 **/
static void printhelp() {
  printf("USAGE: gnunetd [OPTIONS]\n"\
	 "-h, --help                    : "\
	 "print this page\n"\
	 "-c FILENAME, --config=FILENAME: "\
	 "load config file (defaults: %s)\n"\
	 "-v, --version                 : "\
	 "print the version number\n"\
	 "-u USER, --user=USER          : "\
	 "run as user USER\n"\
	 "-L LOGLEVEL, --loglevel=LOGL  : "\
	 "set the loglevel\n"\
	 "-d, --debug                   : "\
	 "run in debug mode, do not detach from console\n"\
	 "-l, --livedot                 : "\
	 "print a dot every second to show that gnunetd is alive\n",
	 DEFAULT_CONFIG_FILE);
}


/**
 * Perform option parsing from the command line. 
 **/
int parseCommandLine(int argc, 
		     char * argv[]) {
  int cont = OK;
  int c;
  struct passwd * pws;

  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      { "loglevel",1, 0, 'L' },
      { "config",  1, 0, 'c' },
      { "version", 0, 0, 'v' },
      { "help",    0, 0, 'h' },
      { "user",    1, 0, 'u' },
      { "debug",   0, 0, 'd' },
      { "livedot", 0, 0, 'l' },
      { "padding", 1, 0, 'p' },
      { 0,0,0,0 }
    };
    
    c = getopt_long(argc,
		    argv, 
		    "vhdc:u:L:lp:", 
		    long_options, 
		    &option_index);
    
    if (c == -1) 
      break;  // No more flags to process
    
    switch(c) {
    case 'p':
      freeNonNull(setConfigurationString("GNUNETD-EXPERIMENTAL",
					 "PADDING",
					 optarg),
		  "startup: option parsing p"); 
      break;      
    case 'l':
      addCronJob(&printDot,
		 1 * cronSECONDS, 
		 1 * cronSECONDS, 
		 NULL);
      break;
    case 'c': 
      freeNonNull(setConfigurationString("FILES",
					 "gnunet.conf",
					 optarg),
		  "startup: option parsing c"); 
      break;
    case 'v': 
      printf("GNUnet v%s\n",
	     VERSION);
      cont = SYSERR;
      break;
    case 'h': 
      printhelp(); 
      cont = SYSERR;
      break;
    case 'L':
      freeNonNull(setConfigurationString("GNUNETD",
					 "LOGLEVEL",
					 optarg),
		  "startup option parsing L");
      break;
    case 'd':
      debug_flag = YES;
      setConfigurationString("GNUNETD",
			     "LOGFILE",
			     NULL);
      break;
    case 'u':
      pws = getpwnam(optarg);
      if (pws == NULL) {
        LOG(LOG_WARNING, 
	    "WARNING: User %s not known, can not change UID to it. Will continue...",
	    optarg);
        break;
      }
      if ( (0 != setregid(pws->pw_gid, pws->pw_gid)) || 
	   (0 != setreuid(pws->pw_uid, pws->pw_uid)) )
	LOG(LOG_WARNING, 
	    "WARNING: Can not change user/group to %s. Will continue...",
	    optarg);
      break;
    default:
      LOG(LOG_FAILURE, 
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      cont = SYSERR;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    LOG(LOG_WARNING, 
	"WARNING: Invalid arguments: ");
    while (optind < argc)
      LOG(LOG_WARNING, 
	  "%s ", argv[optind++]);
    LOG(LOG_FATAL,
	"FATAL: Invalid arguments. Exiting.\n");
    return SYSERR;
  }
  return cont;
}


/* end of startup.c */
