/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/** 
 * @file util/statuscalls.c
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * Status calls implementation for load management.
 *
 * Todo:
 * - add capability to be notified by client-code that
 *   storage allocation has changed for (allocated, freed).
 * - determining load between calls to /proc might be made
 *   interface specific (and the whole policy/economics/load
 *   handling should be carefully investigated for
 *   additional deficiencies - IW).
 **/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "util/statuscalls.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "util/configuration.h"
#include "util/semaphore.h"
#include "util/cron.h"
#include "util/statistics.h"

/* where to read network interface information from
   under Linux */
#define PROC_NET_DEV "/proc/net/dev"

typedef struct {
  unsigned long long last_in;
  unsigned long long last_out;
} NetworkStats;

static char ** interfacePtrs; /* Pointer to the name of each interface, has numInterfaces entries */
static int numInterfaces = 0; /* how many interfaces do we have? */

/* configuration */
static int maxNetDownBPS;
static int maxNetUpBPS;
static int maxNetLoad; 		   /* in percent of maxNetXXXBPS */
static int maxCPULoad; 		   /* in percent of 1 CPU */
static int lastNetResultUp = -1;   /* the max upstream load we saw last time */
static int lastNetResultDown = -1; /* the max dnstream load we saw last time */
static cron_t lastnettime   =  0;  /* when did we check last time? */
static int useBasicMethod = YES;   /* how to measure traffic */

/* tracking */
static char ** tracked_files = NULL;
static int tracked_file_count = 0;
static NetworkStats * last_net_results; /* has numInterfaces entries */
static long long spaceLeft; /* in KB */
static long long spaceUsed; /* in KB */
static DriveFullCallback * driveFullCallbacks;
static int dfc_count;

/* quota */
static unsigned int diskQuota; /* in MB! */
static int stat_handle_kb_used;
static int stat_handle_kb_remaining;
static int stat_handle_network_load;
static int stat_handle_cpu_load;

/* for async configuration update! */
static Mutex statusMutex;

/* amount of kbs gnunetd tries to keep free at all times */
#define MIN_FREE 1024

/**
 * Flag to ensure that we do not start a
 * second job while we are already
 * cleaning up...
 **/
static Semaphore * signalSemaphore = NULL;

static NetworkStats globalTrafficBetweenProc;

/**
 * Increment the number of bytes sent. Transports should use this
 * so that statuscalls module can measure gnunet traffic usage between 
 * calls to /proc.
 *
 * Fixme: the caller doesn't know what interface it is attached to,
 * so this is currently global.
 **/
void incrementBytesSent(unsigned long long delta) {
  MUTEX_LOCK(&statusMutex);
  globalTrafficBetweenProc.last_out += delta;
  MUTEX_UNLOCK(&statusMutex);
}

void incrementBytesReceived(unsigned long long delta) {
  MUTEX_LOCK(&statusMutex);
  globalTrafficBetweenProc.last_in += delta;
  MUTEX_UNLOCK(&statusMutex);
}

static void spaceJob() {
  int i;
  long long unitsFreed;
  long long oldNeed;
  long long needToFree;
 
  needToFree = spaceLeft - MIN_FREE;
  oldNeed = needToFree;
  for (i=0;i<dfc_count;i++) 
    driveFullCallbacks[i](&needToFree);
  unitsFreed = needToFree - oldNeed;
 
  spaceUsed -= unitsFreed;
  spaceLeft += unitsFreed;

  statSet(stat_handle_kb_used,
          (spaceUsed > 0 ? spaceUsed : 0));
  statSet(stat_handle_kb_remaining,
	  (spaceLeft > 0 ? spaceLeft : 0));

  for (i=0;i<dfc_count+1;i++)
    semaphore_up(signalSemaphore);
  MUTEX_LOCK(&statusMutex);
  free_semaphore(signalSemaphore);
  signalSemaphore = NULL;
  MUTEX_UNLOCK(&statusMutex);
}

static void cronSpaceMonitor() {
  unsigned long long use;
  int i;

  LOG(LOG_CRON,
      "CRON: cronSpaceMonitor\n");
  MUTEX_LOCK(&statusMutex);
  use = 0;
  /* FIXME: deletion from gdb/tdb does not necessary affect getFileSize()
   * at all. After hitting the limit for the first time, we will use spaceJob() 
   * each time cronSpaceMonitor() gets called, before the db is reorganized 
   * (very costly). This might call for redesign. 
   */
  for (i=0;i<tracked_file_count;i++)
    use += getFileSize(tracked_files[i]);
  spaceUsed = use / 1024;
  spaceLeft /* KB */
    = (diskQuota /* MB */ * 1024) - (use /* bytes */ / 1024);
  statSet(stat_handle_kb_used,
	  (spaceUsed > 0 ? spaceUsed : 0));
  statSet(stat_handle_kb_remaining,
          (spaceLeft > 0 ? spaceLeft : 0));

  if (spaceLeft < MIN_FREE) {
    if (signalSemaphore == NULL) {
      pthread_t job;

      signalSemaphore = new_semaphore(0);
      PTHREAD_CREATE(&job,
		     (void * (*)()) &spaceJob, 
		     NULL,
		     1024*64);
      pthread_detach(job);
    }
  }
  MUTEX_UNLOCK(&statusMutex);
  LOG(LOG_CRON,
      "CRON: cronSpaceMonitor done.\n");
}

void increaseKilobyteUsed(int kb) {
  MUTEX_LOCK(&statusMutex);
  spaceLeft -= kb;
  if (spaceLeft < MIN_FREE)
    addCronJob((CronJob)&cronSpaceMonitor,
	       0,
	       0,
	       NULL);
  MUTEX_UNLOCK(&statusMutex);
}

/**
 * Register a callback for drive-full monitoring
 **/
int registerDriveFullCallback(DriveFullCallback callback) {
  DriveFullCallback * tmp;

  MUTEX_LOCK(&statusMutex);
  tmp = xmalloc((dfc_count+1)*sizeof(DriveFullCallback),
		"growing DFC array");
  if (driveFullCallbacks != NULL) {
    memcpy(tmp,
	   driveFullCallbacks,
	   dfc_count * sizeof(DriveFullCallback));
    xfree(driveFullCallbacks,
	  "growing DFC array");
  }
  driveFullCallbacks = tmp;  
  driveFullCallbacks[dfc_count++] = callback;
  MUTEX_UNLOCK(&statusMutex);
  return OK;
}

/**
 * Register a callback for drive-full monitoring
 **/
int unregisterDriveFullCallback(DriveFullCallback callback) {
  int i;

  MUTEX_LOCK(&statusMutex);
  if (signalSemaphore != NULL)
    semaphore_down(signalSemaphore);
  
  for (i=0;i<dfc_count;i++)
    if (driveFullCallbacks[i] == callback) {
      driveFullCallbacks[i] = driveFullCallbacks[--dfc_count];
      MUTEX_UNLOCK(&statusMutex);
      return OK;
    }
  MUTEX_UNLOCK(&statusMutex);
  return SYSERR;
}
 
static void cronLoadUpdate(void * unused) {
  getCPULoad();
  getNetworkLoad();
}

/**
 * The following method is called in order to initialize the status calls
 * routines. After that it is safe to call each of the status calls separately
 * @return OK on success and SYSERR on error (or calls errexit).
 **/
void initStatusCalls() {
  LOG(LOG_DEBUG,
      "DEBUG: in initStatusCalls\n");
  MUTEX_CREATE_RECURSIVE(&statusMutex);
  tracked_file_count = 0;
  last_net_results = NULL; /* has numInterfaces entries */
  spaceLeft = 0; /* in KB */
  spaceUsed = 0; /* in KB */
  dfc_count = 0;
  interfacePtrs = NULL;
  last_net_results = NULL;
  tracked_files = NULL;
  driveFullCallbacks = NULL;
  globalTrafficBetweenProc.last_in = 0;
  globalTrafficBetweenProc.last_out = 0;
  stat_handle_kb_used
    = statHandle("# kb of storage used");
  stat_handle_kb_remaining
    = statHandle("# kb of storage remaining");
  stat_handle_network_load
    = statHandle("% of allowed network load");
  stat_handle_cpu_load
    = statHandle("% of allowed cpu load");
  cronTime(&lastnettime);  
  resetStatusCalls();
  networkUsage(); 
  cpuUsage();  
  /* every hour, check space consumption */
  addCronJob(&cronSpaceMonitor,
	     10 * cronMINUTES,
	     10 * cronMINUTES,
	     NULL);
  addCronJob(&cronLoadUpdate,
	     10 * cronSECONDS,
	     10 * cronSECONDS,
	     NULL);
}

/**
 * Shutdown the status calls module.
 **/
void doneStatusCalls() {
  delCronJob(&cronSpaceMonitor,
	     1 * cronHOURS,
	     NULL);
  delCronJob(&cronLoadUpdate,
	     10 * cronSECONDS,
	     NULL);
  xfree(interfacePtrs[0], 
	"doneStatusCalls: interface-string(s)");
  xfree(interfacePtrs,
	"doneStatusCalls: interfacePtrs array");
  if (last_net_results != NULL)
    xfree(last_net_results,
	  "doneStatusCalls: network statistics");    
  freeNonNull(tracked_files,
	      "doneStatusCalls: freeing list of tracked files");
  if (driveFullCallbacks != NULL)
    xfree(driveFullCallbacks,
	  "doneStatusCalls: freeing list of out of space callbacks");
  MUTEX_DESTROY(&statusMutex);
}

/**
 * Re-read the configuration for statuscalls.
 **/
void resetStatusCalls() {
  char * interfaces;
  char * ifcs;
  int start;

  MUTEX_LOCK(&statusMutex);
  interfaces 
    = getConfigurationString("LOAD",
			     "INTERFACES");
  ifcs = interfaces;
  /* fail if config-file is incomplete */
  if (interfaces == NULL) 
    errexit("No network interfaces defined!\n");

  /* The string containing the interfaces is formatted in the following way: 
   * each comma is replaced by '\0' and the pointers to the beginning of every 
   * interface are stored
   */
  numInterfaces = 0;
  start = YES;
  while (1) {
    if (*interfaces == '\0') {
      if (start == NO) 
	numInterfaces++;      
      break;
    }
    if ( ((*interfaces>='a') && (*interfaces<='z')) || 
	 ((*interfaces>='A') && (*interfaces<='Z')) || 
	 ((*interfaces>='0') && (*interfaces<='9')) ) {
      start = NO;         
    } else {
      if (*interfaces != ',')
	errexit("interfaces string (%s) invalid\n",
		ifcs);
      if (start == NO) {
	start = YES;
	numInterfaces++;
      }
    }
    interfaces++;
  }
  if (numInterfaces <= 0) 
    errexit("ERROR: No network interfaces specified in he configuration file\n");

  interfacePtrs = xmalloc(sizeof(char*) * numInterfaces,
			  "allocating space for interfaces");
  last_net_results = xmalloc(sizeof(NetworkStats) * numInterfaces,
			     "allocating space for interface statistics");
  memset(last_net_results, 0,
	 sizeof(NetworkStats) * numInterfaces);

  /* 2nd pass, this time remember the positions */
  interfaces = ifcs;
  numInterfaces = 0;
  start = YES;
  while (1) {
    if (*interfaces=='\0') {
      if (start == NO) 
	numInterfaces++;      
      break;
    }
    if ( ((*interfaces>='a') && (*interfaces<='z')) || 
	 ((*interfaces>='A') && (*interfaces<='Z')) || 
	 ((*interfaces>='0') && (*interfaces<='9')) ) {
      if (start == YES) {
	start = NO;   
	interfacePtrs[numInterfaces] = interfaces;	
      }
    } else {
      if (start == NO) {
	start = YES;
	*interfaces = '\0';	
	numInterfaces++;
      }
    }
    interfaces++;
  }
  
  useBasicMethod 
    = testConfigurationString("LOAD",
 			      "BASICLIMITING",
			      "YES");
  maxNetDownBPS 
    = getConfigurationInt("LOAD",
			  "MAXNETDOWNBPSTOTAL");
  if (maxNetDownBPS == 0)
    maxNetDownBPS = 50000;
  maxNetUpBPS 
    = getConfigurationInt("LOAD",
			  "MAXNETUPBPSTOTAL");
  if (maxNetUpBPS == 0)
    maxNetUpBPS = 50000;
  maxNetLoad
    = getConfigurationInt("LOAD",
			  "MAXNETLOAD");
  if (maxNetLoad == 0 || useBasicMethod == YES)
    maxNetLoad = 100;
  maxCPULoad
    = getConfigurationInt("LOAD",
			  "MAXCPULOAD");
  if (maxCPULoad == 0)
    maxCPULoad = 100;
  diskQuota
    = getConfigurationInt("LOAD",
			  "DISKQUOTA");
  if (diskQuota == 0)
    diskQuota = 1;
  MUTEX_UNLOCK(&statusMutex);  
  cronSpaceMonitor();
}

/**
 * The basic usage meter considers only gnunetd traffic.
 **/
static int networkUsageBasic() {
  cron_t now, elapsedTime;
  int upUsage;
  int downUsage;
  double upCurrent;
  double downCurrent;
  double currentLoad;
  
  MUTEX_LOCK(&statusMutex);
  /* If called less than 1 seconds ago, calc average 
     between avg. traffic of last second and the additional
     gained traffic. If more than 1 second between calls,
     set the avg bytes/sec as the usage of last second. */
  cronTime(&now);  
  elapsedTime = now - lastnettime;
  if (elapsedTime < cronSECONDS) {
    upUsage = (int)((double)lastNetResultUp + 
	 	    (double)globalTrafficBetweenProc.last_out) /
		    (1.0 + ((double)elapsedTime / (double)cronSECONDS));
    downUsage = (int)((double)lastNetResultDown + 
		     (double)globalTrafficBetweenProc.last_in) /
		     (1.0 + ((double)elapsedTime / (double)cronSECONDS));
  } else {
    upUsage = (int)((double)globalTrafficBetweenProc.last_out /
    			       ( (double)elapsedTime / (double)cronSECONDS ));
    downUsage = (int)((double)globalTrafficBetweenProc.last_in /
    			       ( (double)elapsedTime / (double)cronSECONDS ));
    lastNetResultUp = upUsage;
    lastNetResultDown = downUsage;
    globalTrafficBetweenProc.last_in = 0;
    globalTrafficBetweenProc.last_out = 0;
    lastnettime = now;
  }
   
  upCurrent = (double)upUsage / (double)maxNetUpBPS;
  downCurrent = (double)downUsage / (double)maxNetDownBPS;

  if(upCurrent > downCurrent)
    currentLoad = upCurrent;
  else
    currentLoad = downCurrent;
 
/*
  fprintf(stderr, "DEBUG: incrB gti=%llu gto=%llu ",
        globalTrafficBetweenProc.last_in, globalTrafficBetweenProc.last_out);
  fprintf(stderr, " e=%d d=%d u=%d load=%e ret=%d\n", 
	(int)(elapsedTime),
    	downUsage, upUsage, currentLoad, (int)(100.0 * currentLoad));
*/

  MUTEX_UNLOCK(&statusMutex);
  return (int)(100.0 * currentLoad);
}

/**
 * The advanced usage meter takes into account all network traffic.
 * This might be problematic on systems where the same interface 
 * can have different capabilities for different types of traffic 
 * (like support for very fast local traffic but capable of 
 * handling only small-scale inet traffic).
 **/
static int networkUsageAdvanced() {
#define MAX_PROC_LINE 5000
  int retval = -1;
  char line[MAX_PROC_LINE];
  unsigned long long rxnew, txnew;
  unsigned long long rxdiff, txdiff;
  int i=0;
  int ifnum;
  cron_t now, elapsedtime;
  
#ifdef LINUX
  FILE * proc_net_dev;
  char * data;
#else
  FILE * command;
#endif

  MUTEX_LOCK(&statusMutex);
  /* If we checked /proc less than 2 seconds ago, don't do
     it again, but add internal gnunet traffic increments */
  cronTime(&now);  
  elapsedtime = now - lastnettime;
  if (elapsedtime < 2 * cronSECONDS) {
    int upDelta;
    int downDelta;
    int newTempMax;

    if (elapsedtime < 1*cronSECONDS)   /* reqd. for delta calc. below */
      elapsedtime = 1*cronSECONDS;

    upDelta = (100 * globalTrafficBetweenProc.last_out * cronSECONDS) /
	   (elapsedtime * maxNetUpBPS);
    downDelta = (100 * globalTrafficBetweenProc.last_in * cronSECONDS) /
	   (elapsedtime * maxNetDownBPS);

    /* be a bit conservative, think the load measured from /proc
       as current base load and add internal gnunet load to it.
       any better ideas? perhaps we should always average the load
       of past 2 secs or something... */
    if (lastNetResultUp + upDelta > lastNetResultDown + downDelta)
      newTempMax = lastNetResultUp + upDelta;
    else
      newTempMax = lastNetResultDown + downDelta;
    /*
    fprintf(stderr, "DEBUG: incrB gti=%llu gto=%llu ",
        globalTrafficBetweenProc.last_in, globalTrafficBetweenProc.last_out);
    fprintf(stderr, " e=%d d=%d u=%d newMax=%d\n", 
	(int)(elapsedtime+1),
    	downDelta, upDelta, newTempMax);
    */

    MUTEX_UNLOCK(&statusMutex);
    return newTempMax;
  } else {
    globalTrafficBetweenProc.last_in = 0;
    globalTrafficBetweenProc.last_out = 0;
  }
  lastnettime = now;

  /* ok, full program... */
  if ( (maxNetUpBPS == 0) || 
       (maxNetDownBPS == 0) ) {
    lastNetResultUp = -1;
    lastNetResultDown = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  rxdiff = 0;
  txdiff = 0;
  
#ifdef LINUX
  proc_net_dev = fopen(PROC_NET_DEV, "r");
  /* Try to open the file*/
  if (NULL == proc_net_dev) {
    LOG(LOG_ERROR,
	"ERROR: Could not open %s (%s)\n", 
	PROC_NET_DEV,
	strerror(errno));
    MUTEX_UNLOCK(&statusMutex);
    lastNetResultUp = -1;
    lastNetResultDown = -1;
    return -1;
  }
  ifnum = 0;
  /* Parse the line matching the interface ('eth0') */
  while ( (!feof(proc_net_dev)) && 
	  ( ifnum < numInterfaces) ) {    
    fgets(line,
	  MAX_PROC_LINE, 
	  proc_net_dev);
    
    for (i=0;i<numInterfaces;i++) {
      if (NULL != strstr(line, interfacePtrs[i]) ) {
	data = (char*)strchr(line, ':');
	data++;	
	if (sscanf(data,
		   "%llu %*s %*s %*s %*s %*s %*s %*s %llu",
		  &rxnew, &txnew) != 2) {
	  fclose(proc_net_dev);
	  errexit("ERROR: reading interface data from %s\n",
		  PROC_NET_DEV);
	} 	
	rxdiff += rxnew - last_net_results[ifnum].last_in;	  
	txdiff += txnew - last_net_results[ifnum].last_out;
	/*LOG(LOG_NOTHING,
	    "NOTHING: (%d<%d) %s -- rxdiff %llu = %llu - %llu; txdiff %llu = %llu - %llu\n",
	    ifnum, numInterfaces,
	    line,
	    rxdiff, rxnew, last_net_results[ifnum].last_in,
	    txdiff, txnew, last_net_results[ifnum].last_out);*/
	last_net_results[ifnum].last_in = rxnew;
	last_net_results[ifnum].last_out = txnew;
	ifnum++;
	break;
      }
    }
  }
  fclose(proc_net_dev);
#else
  if (1) {
    MUTEX_UNLOCK(&statusMutex);
    return 0;
  }
  if ( ( command = popen("netstat -n -f inet -i", "r") ) == NULL ) {
    LOG(LOG_ERROR,
	"ERROR: could not open netstat\n");
    lastNetResultUp = -1;
    lastNetResultDown = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  ifnum = 0;
  while ( (!feof(command)) && 
	  (ifnum < numInterfaces ) ) {
    fgets(line, 
	  MAX_PROC_LINE, 
	  command);
    for (i=0; i < numInterfaces; i++) {
      if ( NULL != strstr(line, interfacePtrs[i]) ) {
	if(sscanf(line, "%*s %*s %*s %*s %llu %*s %llu %*s %*s", 
		  &rxnew, &txnew) != 2 ) {
	  pclose(command);
	  errexit("ERROR: reading interface data using netstat\n");
	}
	rxdiff += rxnew - last_net_results[ifnum].last_in;
	last_net_results[ifnum].last_in = rxnew;
	txdiff += txnew - last_net_results[ifnum].last_out;
	last_net_results[ifnum].last_out = txnew;
	ifnum++;
	break;
      } /* if match */
    }  /* for all interfaces, find match */
  } /* while: for all lines in proc */
  pclose(command);
#endif
  
  lastNetResultUp 
    = (100 * txdiff * cronSECONDS) / (elapsedtime * maxNetUpBPS);
  lastNetResultDown 
    = (100 * rxdiff * cronSECONDS) / (elapsedtime * maxNetDownBPS);
	
  if (lastNetResultUp > lastNetResultDown)
    retval = lastNetResultUp;
  else
    retval = lastNetResultDown;  

  MUTEX_UNLOCK(&statusMutex);
  return(retval);  
}

/**
 * The following routine returns the percentage of available used
 * bandwidth.  Example: If 81 is returned this means that 81% of the
 * network bandwidth of the host is consumed.  The method
 * initStatusCalls() should be called before this routine is invoked.
 * If there is an error the method returns -1.
 **/
int networkUsage() {
  if ( useBasicMethod == YES )
    return networkUsageBasic();
  else
    return networkUsageAdvanced();
}


/**
 * The following routine returns a number between 0-100 (can be larger than 100
 * if the load is > 1) which indicates the percentage CPU usage.  
 * 
 * Before its first invocation the method initStatusCalls() must be called.
 * If there is an error the method returns -1
 **/
int cpuUsage(){
  cron_t now, elapsedtime;
  static cron_t lastcputime  =  0; /* In seconds */
  static int lastcpuresult = -1;
#ifndef HAVE_GETLOADAVG
#ifdef LINUX
  int ret = -1;
  static int last_cpu_results[4] = { 0, 0, 0, 0 };
  FILE* proc_stat;
  char line[128];
  int user_read, system_read, nice_read, idle_read;
  int user, system, nice, idle;
  int usage_time=0, total_time=1;
#endif
#else
  double loadavg;
#endif

  MUTEX_LOCK(&statusMutex);
  cronTime(&now);  
  elapsedtime = now - lastcputime;
  if (elapsedtime < 2 * cronSECONDS) {
    MUTEX_UNLOCK(&statusMutex);
    return lastcpuresult;
  }
  lastcputime = now;
  
#ifndef HAVE_GETLOADAVG
#ifdef LINUX
  /* Get the current time in microseconds */

  proc_stat = fopen("/proc/stat", "r");
  /* Try to open the file*/
  if (NULL == proc_stat) {
    LOG(LOG_ERROR,
	"ERROR: Could not open /proc/stat");
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  /* Get the first line with the data */
  if(fgets(line, 128, proc_stat)==NULL) {
    LOG(LOG_ERROR,
	"ERROR: reading /proc/stat file");
    fclose(proc_stat);
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  if (sscanf(line, "%*s %i %i %i %i", 
	     &user_read, &system_read, &nice_read,
	     &idle_read) != 4) {
    fclose(proc_stat);
    LOG(LOG_ERROR,
	"ERROR: decoding /proc/stat file");
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }

  /* Store the current usage*/
  user = user_read - last_cpu_results[0];
  system = system_read - last_cpu_results[1];
  nice = nice_read - last_cpu_results[2];
  idle = idle_read - last_cpu_results[3];
  
  /* Calculate the % usage */
  if ((user + system + nice + idle) > 0) {
    usage_time = user + system + nice;
    total_time = usage_time + idle;
  }
  ret = (100 * usage_time) / total_time;

  /* Store the values for the next calculation*/
  last_cpu_results[0] = user_read;
  last_cpu_results[1] = system_read;
  last_cpu_results[2] = nice_read;
  last_cpu_results[3] = idle_read;
  fclose(proc_stat);

  lastcpuresult = ret;
  MUTEX_UNLOCK(&statusMutex);
  return ret;
#else
  /* loadaverage not defined and not linux: error */
  lastcpuresult = -1;
  MUTEX_UNLOCK(&statusMutex);
  return -1;
#endif
#else
  if ( getloadavg(&loadavg, 1) == -1 ) {
    LOG(LOG_ERROR,
	"ERROR: getting load average\n");
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  lastcpuresult = (int) (100 * loadavg);
  MUTEX_UNLOCK(&statusMutex);
  return lastcpuresult;
#endif
}

/**
 * Get the load of the network relative to what is allowed.
 * @return the network load as a percentage of allowed
 *        (100 is equivalent to full load)
 **/
int getNetworkLoad() {
  static int lastRet = 0;
  int ret = (100 * networkUsage()) / maxNetLoad;
  if (ret > lastRet) 
    lastRet = ret; /* quickly rise to higher loads */
  else {
    ret = (ret + 7 * lastRet)/8;
    lastRet = ret; /* but only slowly go back! */
  }
  statSet(stat_handle_network_load, ret);
  return ret;
}

/**
 * Get the load of the CPU relative to what is allowed.
 * @return the CPU load as a percentage of allowed
 *        (100 is equivalent to full load)
 **/
int getCPULoad() {
  static int lastRet = 0;
  int ret = (100 * cpuUsage()) / maxCPULoad;
  if (ret > lastRet) 
    lastRet = ret; /* quickly rise to higher loads */
  else {
    ret = (ret + 7 * lastRet)/8;
    lastRet = ret; /* but only slowly go back! */
  }
  statSet(stat_handle_cpu_load, ret);
  return ret;
}

/**
 * @return the number of kilobytes GNUnet is allowed to use.
 **/
unsigned long long getDiskQuota() {
  return diskQuota * 1024;
}

/**
 * @return the number of kilobytes we have left.
 **/
long long getSpaceLeft() {
  return spaceLeft;
}

/**
 * Register a file for monitoring space consumption.
 **/
int registerSpaceMonitor(char * filename) {
  char ** tmp;
  unsigned long long size;

  LOG(LOG_DEBUG,
      "DEBUG: space monitor for %s registered\n",
      filename);
  MUTEX_LOCK(&statusMutex);
  tmp = xmalloc((tracked_file_count+1) * sizeof(char*),
		"registerSpaceMonitor: growing file tracking list");
  if (tracked_file_count > 0) {
    memcpy(tmp,
	   tracked_files,
	   sizeof(char*) * tracked_file_count);
    xfree(tracked_files,
	  "registerSpaceMonitor: growing");
  }
  tracked_files = tmp;
  tracked_files[tracked_file_count++] = STRDUP(filename);

  size = getFileSize(filename) / 1024;
  spaceLeft -= size;
  spaceUsed += size;
  statSet(stat_handle_kb_used,
	  (spaceUsed > 0 ? spaceUsed : 0));
  statSet(stat_handle_kb_remaining,
          (spaceLeft > 0 ? spaceLeft : 0)); /* <0 not allowed in stats! */
  MUTEX_UNLOCK(&statusMutex);
  return OK;
}

/* end of statusCalls.c */
